package com.flyfish.framework.service;

import com.flyfish.framework.config.constants.UserCacheKeys;
import com.flyfish.framework.configuration.jwt.TokenProvider;
import com.flyfish.framework.domain.AdminUserDetails;
import com.flyfish.framework.domain.base.IUser;
import com.flyfish.framework.enums.UserStatus;
import com.flyfish.framework.utils.Assert;
import com.flyfish.framework.utils.ReactiveRedisOperations;
import lombok.RequiredArgsConstructor;
import org.apache.commons.lang3.BooleanUtils;
import org.springframework.security.authentication.*;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.AuthenticationException;
import org.springframework.security.core.context.ReactiveSecurityContextHolder;
import org.springframework.security.core.context.SecurityContext;
import org.springframework.security.core.context.SecurityContextImpl;
import org.springframework.security.core.userdetails.UserDetails;
import org.springframework.security.core.userdetails.UsernameNotFoundException;
import org.springframework.security.crypto.password.NoOpPasswordEncoder;
import org.springframework.stereotype.Service;
import org.springframework.web.server.ServerWebExchange;
import reactor.core.publisher.Mono;

import javax.annotation.PostConstruct;
import javax.annotation.Resource;
import java.util.HashMap;
import java.util.Map;
import java.util.function.Function;
import java.util.function.Supplier;

/**
 * 基于mongo实现的用户详情
 *
 * @author wangyu
 */
@RequiredArgsConstructor
@Service
public class MongoUserDetailsServiceImpl implements MongoUserDetailsService {

    // 存储用户校验规则的map
    private static final Map<Function<IUser, Boolean>, Supplier<AuthenticationException>> checkMap;

    static {
        checkMap = new HashMap<>();
        // 初始化用户校验规则
        checkMap.put(user -> null != user.getEnable() && !user.getEnable() ||
                user.getStatus() == UserStatus.DISABLED, () -> new DisabledException("用户被禁用"));
        checkMap.put(user -> user.getStatus() == UserStatus.LOCKED,
                () -> new LockedException("账户已经锁定！请联系管理员修改密码！"));
    }

    private final UserService userService;
    private final UserDetailsConverter userDetailsConverter;

    @Resource
    private TokenProvider tokenProvider;
    @Resource
    private ReactiveRedisOperations reactiveRedisOperations;

    private ReactiveAuthenticationManager authenticationManager;


    @PostConstruct
    @SuppressWarnings("deprecation")
    private void init() {
        UserDetailsRepositoryReactiveAuthenticationManager manager =
                new UserDetailsRepositoryReactiveAuthenticationManager(this);
        manager.setPasswordEncoder(NoOpPasswordEncoder.getInstance());
        authenticationManager = manager;
    }

    @Override
    public Mono<UserDetails> updatePassword(UserDetails userDetails, String password) {
        return userService.updateById(((IUser) userDetails).toUser()).thenReturn(userDetails);
    }

    @Override
    public Mono<UserDetails> findByUsername(String username) {
        String key = UserCacheKeys.get(username);
        // 优先从缓存读取，如果缓存不存在，查询转换后放入缓存
        return reactiveRedisOperations.hasKey(key).flatMap(exists -> Boolean.TRUE.equals(exists) ?
                        reactiveRedisOperations.get(key) :
                        userService.findByUsername(username)
                                .flatMap(userDetailsConverter::mapToUserDetails)
                                .flatMap(detail -> reactiveRedisOperations.set(key, detail).thenReturn(detail))
                )
                .flatMap(this::validate)
                .switchIfEmpty(Mono.defer(() -> Mono.error(new UsernameNotFoundException("用户不存在！"))));
    }

    /**
     * 校验
     *
     * @param details 用户
     * @return 结果
     */
    private Mono<UserDetails> validate(UserDetails details) {
        if (details instanceof AdminUserDetails) {
            AdminUserDetails user = (AdminUserDetails) details;
            return checkMap.entrySet().stream()
                    .filter(entry -> entry.getKey().apply(user))
                    .findFirst()
                    .map(entry -> Mono.<UserDetails>error(entry.getValue().get()))
                    .orElse(Mono.just(user));
        }
        return Mono.just(details);
    }

    /**
     * 通过用户名密码直接认证
     *
     * @param user 用户
     * @return 结果
     */
    @Override
    public Mono<Authentication> authenticate(UserDetails user, ServerWebExchange exchange) {
        return loadContext(user)
                .flatMap(securityContext -> Mono.fromCallable(() -> {
                            // 添加jwtToken
                            tokenProvider.addToken(exchange, securityContext.getAuthentication());
                            return Mono.empty();
                        })
                        .contextWrite(ReactiveSecurityContextHolder.withSecurityContext(
                                Mono.just(securityContext)))
                        .then(Mono.just(securityContext)))
                .map(SecurityContext::getAuthentication)
                .doOnNext(authentication -> {
                    if (authentication.getPrincipal() instanceof IUser) {
                        IUser userDetail = (IUser) authentication.getPrincipal();
                        Assert.isTrue(BooleanUtils.isTrue(userDetail.getApp()), "用户没有移动端登录权限！");
                    }
                });
    }

    @Override
    public Mono<SecurityContext> loadContext(UserDetails user) {
        // 构建账号信息
        UsernamePasswordAuthenticationToken token = new UsernamePasswordAuthenticationToken(
                user.getUsername(), user.getPassword());
        // 登录先
        return authenticationManager.authenticate(token)
                .map(authentication -> {
                    SecurityContextImpl securityContext = new SecurityContextImpl();
                    securityContext.setAuthentication(authentication);
                    return securityContext;
                });
    }

    /**
     * 退出登录
     *
     * @param exchange 数据请求
     * @return 结果
     */
    @Override
    public Mono<Void> logout(ServerWebExchange exchange) {
        return tokenProvider.removeToken(exchange);
    }

}
