package com.flyfish.framework.service;

import com.flyfish.framework.domain.AdminUserDetails;
import com.flyfish.framework.domain.base.IUser;
import com.flyfish.framework.domain.po.Department;
import com.flyfish.framework.domain.po.Role;
import com.flyfish.framework.enums.UserType;
import com.flyfish.framework.utils.CopyUtils;
import com.flyfish.framework.utils.DepartUtils;
import lombok.RequiredArgsConstructor;
import org.apache.commons.collections4.CollectionUtils;
import org.springframework.security.core.userdetails.UserDetails;
import org.springframework.stereotype.Service;
import reactor.core.publisher.Mono;

import java.util.*;
import java.util.stream.Collectors;
import java.util.stream.Stream;

@Service
@RequiredArgsConstructor
public class UserDetailsConverter {

    private final DepartmentService departmentService;

    /**
     * 将用户映射为用户详情，查询所有权限code
     *
     * @param user 用户信息
     * @return 结果
     */
    public Mono<UserDetails> mapToUserDetails(IUser user) {
        AdminUserDetails userDetail = new AdminUserDetails();
        CopyUtils.copyProps(user, userDetail);
        userDetail.setAuthorityCodes(Collections.singleton(Department.PUBLIC));
        if (user.getType() == UserType.SUPER_ADMIN) {
            userDetail.setAuthorityCodes(null);
        } else if (CollectionUtils.isNotEmpty(user.getDepartments())) {
            return judgeDeparts(user, userDetail);
        }
        return Mono.just(userDetail);
    }

    /**
     * 判断部门包含关系，默认看不到本部门的数据却可以看到子部门的
     *
     * @param user        用户信息
     * @param userDetails 用户详情
     */
    private Mono<UserDetails> judgeDeparts(IUser user, AdminUserDetails userDetails) {
        // 整合数据权限
        Set<Role.Authority> authorities = user.getRoles().stream()
                .flatMap(role -> null == role.getAuthorities() ? Stream.empty() : role.getAuthorities().stream())
                .collect(Collectors.toSet());
        // 根据权限组装查询条件
        Set<String> departs = DepartUtils.mergeDeparts(user.getDepartments());
        // 查询所有子部门id
        return departmentService.getSubCodes(departs)
                .map(codes -> {
                    // 取出权限，便于判定
                    boolean admin = authorities.contains(Role.Authority.ADMIN);
                    boolean view = authorities.contains(Role.Authority.VIEW);
                    boolean viewChildren = authorities.contains(Role.Authority.VIEW_CHILDREN);
                    // 全部
                    Set<String> all = new HashSet<>();
                    // 拥有查看本部门的权限
                    if (admin || view) {
                        all.addAll(departs);
                    }
                    // 拥有查看所有下级的权限
                    if (admin || viewChildren) {
                        all.addAll(codes);
                    }
                    // 最终设置权限
                    if (CollectionUtils.isNotEmpty(all)) {
                        userDetails.setAuthorityCodes(union(all, userDetails.getAuthorityCodes()));
                    }
                    userDetails.setVisibleDeparts(union(codes, departs));
                    return userDetails;
                });

    }

    /**
     * 链接两个set
     *
     * @param a 第一个
     * @param b 第二个
     * @return 结果
     */
    private Set<String> union(Set<String> a, Set<String> b) {
        Set<String> result = new HashSet<>();
        result.addAll(a);
        result.addAll(b);
        return result;
    }
}
